import os
import compiler
from bike.parsing.load import load as loadast
from bike.refactor import renameFunction, renameClass, renameMethod
from bike.parsing.output import save as saveast
from bike.parsing.utils import fqn_rcar, fqn_rcdr

def load(path,progressCallback=None):
    path = os.path.expanduser(path)
    path = os.path.abspath(path)
    context = BRMContext(path,progressCallback)
    return context

class NotAPythonModuleOrPackageException: pass
class CouldntLocateASTNodeFromCoordinatesException: pass

class BRMContext:
    def __init__(self, path, progressCallback):
        self.progressCallback = progressCallback
        self.ast = None
        self.load(path)
        
    def load(self, path):
        if (not path.endswith(".py")) and os.path.isfile(path):
            raise NotAPythonModuleOrPackageException
        self.ast = loadast(path, self.progressCallback, self.ast)

    def _getAST(self):
        return self.ast

    # returns a list of saved filenames
    def save(self):
        savedfiles = saveast(self.ast)
        return savedfiles

    # sets a callback to ask the user about method refs which brm
    # can't deduce the type of. The callback must be callable, and
    # take the following parameters:
    #  - filename
    #  - start coords   (object with integer x and y attributes)
    #  - end coords   (object with integer x and y attributes)
    # (start and end coords enclose the problematic method call)
    def setRenameMethodPromptCallback(self,callback):
        self.promptUserClientCallback = callback



    # fqn = fully qualified name  (a.b.foo.bah)
    def renameClass(self,fqn,newname):
        renameClass.renameClass(self.ast, fqn, newname)

    def renameFunction(self,fqn,newname):
        renameFunction.renameFunction(self.ast, fqn, newname)

    def renameMethod(self,fqn,newname):
        promptcallback = None
        if hasattr(self,"promptUserClientCallback"):
            promptcallback = self._promptUser
        renameMethod.renameMethod(self.ast, fqn, newname, 
                                  promptCallback = promptcallback)

    def _promptUser(self,node,contextNode):
        startcoords = node.getStartCoords()
        endcoords = node.getEndCoords()
        filename = node.sourcenode.filename
        return self.promptUserClientCallback(filename,startcoords,endcoords)


    # filename_path must be absolute
    def renameByCoordinates(self, filename_path, x, y, newname):
        node = self._locateNode(filename_path, x, y)
        if node:
            if isinstance(node, compiler.ast.Function):
                if isinstance(node.parentScope, compiler.ast.Class):
                    self.renameMethod(node.fqn,newname)
                else:
                    self.renameFunction(node.fqn,newname)
            if isinstance(node, compiler.ast.Class):
                self.renameClass(node.fqn, newname)
        else:
            raise CouldntLocateASTNodeFromCoordinatesException()

        
    # filename must be an absolute path
    def reloadSourceFile(self, filename):
        top = self.ast
        class SourceVisitor:
            found = 0
            def visitSource(self, node):
                if node.filename == filename:
                    f = open(filename)
                    src = f.read()
                    f.close()
                    node.reload(src)
                    self.found = 1
        return compiler.walk(self.ast, SourceVisitor()).found
        
    def _locateNode(self, filename, x, y):
        class SourceVisitor:
            def visitSource(self, node):
                if node.filename == filename:
                    node.tokenize()
                    self.visit(node.module)            

            def visitFunction(self, node):
                ycoord = node.parsetree[0][2]
                if ycoord == y:
                    tokens = node.getNodeTokens()
                    xstart = tokens[1].start[1]
                    xend = tokens[1].end[1]
                    if x >= xstart and x < xend:
                        self.result = node
                        return
                for child in node.getChildNodes():
                    self.visit(child)
                    

            def visitClass(self, node):
                ycoord = node.parsetree[0][2]
                if ycoord == y:
                    tokens = node.getNodeTokens()
                    xstart = tokens[1].start[1]
                    xend = tokens[1].end[1]
                    if x >= xstart and x < xend:
                        self.result = node
                        return
                for child in node.getChildNodes():
                    self.visit(child)

        try:
            return compiler.walk(self.ast, SourceVisitor()).result
        except AttributeError: # didn't find it
            return None


        
def _deducePackageOfFile(filename):
    package = ""
    dot = ""
    dir = os.path.dirname(filename)
    while dir != "" and \
          os.path.exists(os.path.join(dir,"__init__.py")):
        dir,dirname = os.path.split(dir)
        package = dirname + dot + package
        dot = "."
    return package
