#!/usr/bin/env python 
import setpath

import unittest

from tokenutils import getTokenList

from test_addtokens_support import*

from test_addtokens_support import MockAddTokens_codeIsInlined

import brmtransformer
from brmtransformer import parse

from bike.testutils import*

import testdata

from bike import testdata as globaltestdata


# tests that whole token tree is retrieved
class TestGetTokens(unittest.TestCase):
    def testGetTokens_TokensReturnedCorrectly(self):
        src = globaltestdata.class_and_method
        ast = createAST(src)
        addTokensToASTFromSrc(ast, src)
        self.assertEquals(ast.getTokens(), getTokenList(src))


testdata_module_class = """
class foo:
    pass
"""

class TestVisitImport(unittest.TestCase):
    def test_takesImportTokens(self):
        src = skipNL(globaltestdata.FunctionRef)
        importstmt = getASTNodeFromSrc(src, "Import")
        assert importstmt.getTokens() == src2Tokens(src)[: 7]

class TestVisitFrom(unittest.TestCase):
    def test_takesFromTokens(self):
        src = skipNL(globaltestdata.FunctionRefImportedViaFrom)
        importstmt = getASTNodeFromSrc(src, "From")
        assert importstmt.getTokens() == src2Tokens(src)[: 7]

class TestVisitPass(unittest.TestCase):

    def visitPassWithSrc(self, src):
        target = getAddTokens(src)
        apass = getASTNodeFromSrc(src, "Pass")
        target.visitPass(apass)
        return target, apass

    def testVisitPass_takesPassTokenAndNewline(self):
        src = skipNL(testdata.pass_nocomment)
        target, apass = self.visitPassWithSrc(src)
        assert(apass.getTokens() == src2Tokens(src)[: 2])
        assert(target.tc.peekNextToken() == src2Tokens(src)[-1])


    def testVisitPass_takesPassTokenAndComment(self):
        src = skipNL(testdata.pass_withcomment)
        target, apass = self.visitPassWithSrc(src)
        assert(apass.getTokens() == src2Tokens(src)[: 3])
        assert(target.tc.peekNextToken() == src2Tokens(src)[-1])

    def testVisitPass_takesPassTokenAndInlineComment(self):
        src = skipNL(testdata.pass_inlinecomment)
        target, apass = self.visitPassWithSrc(src)
        assert(apass.getTokens() == src2Tokens(src)[: 3])
        assert(target.tc.peekNextToken() == src2Tokens(src)[-1])

    def testVisitPass_takesPassTokenAndMultilineCommentButLeavesTokenCommentingNextCodeBlock(self):
        src = skipNL(testdata.pass_withMultilineComment)
        target, apass = self.visitPassWithSrc(src)
        assert apass.getTokens() == src2Tokens(src)[:-4]

    def testVisitPass_takesPassTokenWithNoCommentAndLeavesTokenCommentingNextCodeBlock(self):
        src = skipNL(testdata.pass_commentonnextline)
        target, apass = self.visitPassWithSrc(src)
        assert apass.getTokens() == src2Tokens(src)[:-4]

    def testVisitPass_takesPassTokenAndMultilineCommentAndSucceedingBlankLinesButLeavesTokenCommentingNextCodeBlock(self):
        src = skipNL(testdata.pass_withMultilineCommentAndBlankLine)
        target, apass = self.visitPassWithSrc(src)
        assert apass.getTokens() == src2Tokens(src)[:-4]


    def testVisitPass_takesPassTokenAndPreceedingMultilineComment(self):
        src = skipNL(testdata.pass_withpreceedingmultilinecomment)
        target, apass = self.visitPassWithSrc(src)
        assert(apass.getTokens() == src2Tokens(src)[: 4])
        assert(target.tc.peekNextToken() == src2Tokens(src)[-1])



class TestVisitGetattr(unittest.TestCase):
    def testVisitGetattr_addsGetattrTokens(self):
        src = "a.foo"
        target = getAddTokens(src)
        node = getASTNodeFromSrc(src, "Getattr")
        assert node.getTokens() == src2Tokens(src)[:-1]


    def testVisitGetattr_picksOutGetattrTokens(self):
        src = skipNL(globaltestdata.functionContaingObjectCreationAndMethodReferenceWithImport)
        target = getAddTokens(src)
        ast = createAST(src)
        assert getASTNode(ast, "Getattr", 2).getNodeTokens() == src2Tokens(src)[-8:-5]


    def testVisitGetattr_dividesTokensCorrectlyInCompoundGetattrs(self):
        src = "a.foo.bah"
        ast = createAST(src)
        assert getASTNode(ast, "Getattr", 1).getTokens() == src2Tokens(src)[: 3]
        assert getASTNode(ast, "Getattr", 0).getTokens() == src2Tokens(src)[:-1]


class TestVisitName(unittest.TestCase):
    def test_nameTakesItsNameToken(self):
        src = "a.foo.bah"
        ast = createAST(src)
        assert getASTNode(ast, "Name", 0).getNodeTokens() == [src2Tokens(src)[0]]

    def test_GetsNodeIfThereArePreceedingNodes(self):
        src = "raise foobah"
        ast = createAST(src)
        assert getASTNode(ast, "Name").getNodeTokens() == [src2Tokens(src)[1]]


class TestVisitModule(unittest.TestCase, CompoundNodeTestHelper):

    def getEndMarkerToken(self, src):
        token = src2Tokens(src)[-1]
        assert token.type == tokenize.ENDMARKER
        return[token]

    def testVisitModule_worksWithModuleThatHasNoComment(self):
        src = skipNL(testdata_module_noComment)
        amodule = self.createAndVisitModule_withSrc(src)
        assert(amodule.getTokensBefore() == [])
        assert amodule.getTokensAfter() == self.getEndMarkerToken(src)

    def testVisitModule_takesModuleCommentButLeavesCodeComment(self):
        src = skipNL(testdata_module_withComments)
        amodule = self.createAndVisitModule_withSrc(src)
        assert amodule.getTokensBefore() == src2Tokens(src)[: 2]
        assert amodule.getTokensAfter() == src2Tokens(src)[-2:]

    def testVisitModule_doesntTakeCodeComment(self):
        src = skipNL(testdata_module_withCodeCommentButNoModuleComment)
        amodule = self.createAndVisitModule_withSrc(src)
        assert(amodule.getTokensBefore() == [])
        assert amodule.getTokensAfter() == self.getEndMarkerToken(src)

    def testVisitModule_gobblesBlankLinesPreceedingCode(self):
        src = skipNL(testdata_module_withBlankLines)
        amodule = self.createAndVisitModule_withSrc(src)
        assert(amodule.getTokensBefore() == src2Tokens(src)[: 2])
        assert amodule.getTokensAfter() == self.getEndMarkerToken(src)

    def createAndVisitModule_withSrc(self, src):
        target = getAddTokens(src)
        amodule = getASTNodeFromSrc(src, "Module")
        target.visitModule(amodule)
        return amodule


testdata_module_noComment = """
pass
"""

testdata_module_withComments = """
# this comment documents the module

# this documents the code
pass
#this comment also belongs to the module
"""

testdata_module_withCodeCommentButNoModuleComment = """
# this documents the code
pass
"""


testdata_module_withBlankLines = """


pass
"""


class TestVisitClass(unittest.TestCase, CompoundNodeTestHelper):

    def testVisitClass_WorksWithClassThatHasInlineComment(self):
        src = skipNL(testdata_class_singlecomment)
        self.check_VisitClass_GetsAllTokensExceptPassToken(src)

    def testVisitClass_IncludesTrailingNewlinesButNotTrailingComments(self):
        src = skipNL(testdata_class_trailingnewlines)
        aclass = getASTNodeFromSrc(src, "Class")
        target = MockAddTokens(src2Tokens(src))
        target.visitClass(aclass)
        assert aclass.getTokensAfter() == src2Tokens(src)[8: 10]


    def testVisitClass_WorksWithClassThatDoesntHaveAnInlineComment(self):
        src = skipNL(testdata_class_nocomment)
        self.check_VisitClass_GetsAllTokensExceptPassToken(src)

    def testVisitClass_WorksWithClassThatHasPreceedingComment(self):
        src = skipNL(testdata_class_preceedingcomment)
        self.check_VisitClass_GetsAllTokensExceptPassToken(src)


    def testVisitClass_WorksWithClassAllOnOneLine(self):
        src = skipNL(testdata_class_allononeline)
        target = MockAddTokens_codeIsInlined(src2Tokens(src))
        aclass = getASTNodeFromSrc(src, "Class")
        target.visitClass(aclass)
        assert(aclass.getTokensBefore() == src2Tokens(src)[: 3])
        assert(aclass.getTokensAfter() == src2Tokens(src)[-2:-1]) # newline
        #assert(aclass.getTokensAfter() == [])

    def testVisitClass_doesntBarfWithClassAllOnOneLine(self):
        src = testdata_class_allononeline_with_newline_following
        ast = parse(src)
        addTokensToASTFromSrc(ast, src)


    # i.e. class foo(bah)   - the bah token belongs to child Name ast node
    def testDelegatesBaseClassSpecTokensToNodes(self):
        src = globaltestdata.ClassWithNestedInheritanceSpec
        ast = createAST(src)
        classnode = getASTNode(ast, "Class")
        assert classnode.bases[0].getTokens() == src2Tokens(src)[8: 13]

    def testWorksWithBaseClasses(self):
        src = globaltestdata.ClassWithNestedInheritanceSpec
        ast = createAST(src)
        classnode = getASTNode(ast, "Class")
        assert classnode.getNodeTokens() == src2Tokens(src)[5:-1]

    def test_Wierdsubclassproblem(self):
        top = Top()
        createSourceNodeAt(top, globaltestdata.wierdSubclassProblem, "a.foo")
        createSourceNodeAt(top, globaltestdata.theClassAsAFunction, "a.b.bah")
        addtypeinfo(top)


    # private methods

    def check_VisitClass_GetsAllTokensExceptPassToken(self, src):
        addtokens = getAddTokens(src)
        
        aclass = getASTNodeFromSrc(src, "Class")
        addtokens.visitClass(aclass)
        self.check_CompoundNode_GetsAllTokensExceptPassToken(aclass, src)






testdata_class_allononeline = """
class bar: pass
"""

testdata_class_allononeline_with_newline_following = """
class bar: pass

"""

testdata_class_preceedingcomment = """
# preceeding comment
class bar:
    pass
"""

testdata_class_singlecomment = """
class bar:   # some comment
    pass
"""

testdata_class_trailingnewlines = """
class bar:   # some comment
    pass

#another comment
class zzz: pass
"""

testdata_class_nocomment = """
class bar:
    pass
"""

testdata_class_withbases = """
class foo(bah,baz):
    pass
"""


class TestVisitFunction(unittest.TestCase, CompoundNodeTestHelper):

    def visitFuncWithSrc(self, src):
        fn = getASTNodeFromSrc(src, "Function")
        target = MockAddTokens(src2Tokens(src))
        target.visitFunction(fn)
        return fn


    def testVisitFunction_WorksWithFunctionThatHasNoArgs(self):
        src = skipNL(testdata_func_noargs)
        self.doGenericChecksOnFunction_WithOnlyPassLineInBody(src)

    def testVisitFunction_WorksWithFunctionThatHasArgs(self):
        src = skipNL(testdata_func_args)
        self.doGenericChecksOnFunction_WithOnlyPassLineInBody(src)

    def testVisitFunction_WorksWithFunctionThatHasArgsAndDocstring(self):
        src = skipNL(testdata_func_args_docstring)
        self.doGenericChecksOnFunction_WithOnlyPassLineInBody(src)

    def testVisitFunction_WorksWithFunctionThatHasPreceedingComment(self):
        src = skipNL(testdata_func_preceedingcomment)
        self.doGenericChecksOnFunction_WithOnlyPassLineInBody(src)

    def doGenericChecksOnFunction_WithOnlyPassLineInBody(self, src):
        f = self.visitFuncWithSrc(src)
        self.check_CompoundNode_GetsAllTokensExceptPassToken(f, src)

    def testVisitFunction_getsAllTokensToEndOfFunction(self):
        src = skipNL(globaltestdata.FunctionWithOneElement)
        ast = createAST(src)
        fn = getASTNode(ast, "Function", 0)
        assert fn.getNodeTokens() == src2Tokens(src)[3:-1]

    def test_doesntBarfWithStreamOperator(self):
        src = globaltestdata.StreamToStderr
        ast = parse(src)
        addTokensToASTFromSrc(ast, src)

    def test_doesntBarfWithStreamOperator(self):
        src = globaltestdata.StreamToStderr
        ast = parse(src)
        addTokensToASTFromSrc(ast, src)

    def test_doesntBarfWhenMethodCallSplitOverTwoLines(self):
        # this can cause problems, because the tokenlist contains the
        # NL, but the parse tree doesnt
        src = MethodCallSplitOnNewlineAfterComma
        ast = parse(src)
        addTokensToASTFromSrc(ast, src)

    def test_doesntBorfWhenMethodCallSplitOnNewlineAfterOpenBracket(self):
        src = MethodCallSplitOnNewlineAfterOpenBracket
        ast = parse(src)
        addTokensToASTFromSrc(ast, src)


MethodCallSplitOnNewlineAfterOpenBracket = """
def gui():
    class GUI:
        def __init__(self, window, port=7464):
            threading.Thread(
                target=serve, args=(port, self.ready, self.quit)).start()
"""

MethodCallSplitOnNewlineAfterComma = """
s = _center(foo,
            foo).foo()
"""

testdata_func_noargs = """
def foo():
    pass
"""

testdata_func_args = """
def foo(arg1,arg2):
    pass
"""

testdata_func_args_docstring = """
def foo(arg1=3,):
    \"\"\" this is a doc comment \"\"\"
    pass
"""


testdata_func_preceedingcomment = """
# preceeding comment
def bar():
    pass
"""


if __name__ == "__main__":
    unittest.main()
