#!/usr/bin/env python

from __future__ import nested_scopes
import unittest
import parser
import symbol
import token
import tokenize

import tokenutils

from tokenutils import getTokenList
from tokenutils import TokenConverter
from tokenutils import getTokenConverter
import testdata

class TestTokenEquality(unittest.TestCase):
    #test data
    token =  tokenutils.TokenList.token # less typing below:
    tok = token(tokenize.NAME,"hello",(0,0),(0,0),"hello")
    tok_dup = token(tokenize.NAME,"hello",(0,0),(0,0),"hello")
    not_tok = [
        token(tokenize.OP,"hello",(0,0),(0,0),"hello"),
        token(tokenize.NAME,"hell",(0,0),(0,0),"hello"),
        ]
    

    def test__eq__(self):
        assert(TestTokenEquality.tok == TestTokenEquality.tok_dup)
        for i in TestTokenEquality.not_tok:            
            assert (TestTokenEquality.tok != i),"failed with "+str(i)

    def test__ne__(self):
        assert((TestTokenEquality.tok != TestTokenEquality.tok_dup) == 0) 



class TestTokenList(unittest.TestCase):
    def testTokenList(self):
        # basic test to make sure that list is produced
        tokenlist = getTokenList("class foo")
        assert(tokenlist[0].name == "class")
        assert(tokenlist[0].type == tokenize.NAME)
        assert(tokenlist[1].name == "foo")
        assert(tokenlist[1].type == tokenize.NAME)
        assert(tokenlist[2].type == tokenize.ENDMARKER)
        assert(len(tokenlist) == 3)


    # test to check that comments immediately before a block of code
    # are placed after the previous block DEDENT token
    def testTokenList_commentsBeforeCodeBlockOccurAfterPreviousDedentToken(self):
        t = getTokenList(testdata_reorderComment_class)
        # find dedent
        i = 0
        while t[i].type != token.DEDENT: i += 1   

        # check that dedent comes after comment describing prev fn.
        assert(t[i-1].name == '# previous comment\n')

        # and before the newline and comment describing next class
        assert(t[i+1].type == tokenize.NL)
        assert(t[i+2].name == '# This comment should come after the previous class dedent. It\n')
        




    def testTokenList_commentsDescribingClassMemberOccurAfterIndent(self):
        t = getTokenList(testdata_classWithCommentOnNextLine)
        i=0
        while t[i].type != token.INDENT: i += 1
        # assert comment comes after indent
        assert t[i+1].type == tokenize.COMMENT
        
        

testdata_classWithCommentOnNextLine = """
class foo:
    # hello
    pass
"""

testdata_reorderComment_class = """
class foo:
    pass  # this comment should come before class dedent
          # so should this - it might be a continuation of the
          # previous comment

# This comment should come after the previous class dedent. It
# probably documents the next class or group of classes

# so should this
class bah:
    pass
"""

testdata_reorderComment_function = """
class foo:

    def a(self):
        pass  # this comment should come before function dedent
              # so should this - it's probably be a continuation of
              # the previous comment

    # This comment should come after the previous function dedent. It
    # probably documents the next class or group of classes

    # so should this
    def b(self):
        pass
    
"""


# TODO: These need replacing with tests for getTokens()
class TestTokenConverter(unittest.TestCase):

    def testPeekNextToken_showsTheNextToken(self):
        tlist = getTokenList("class foo: pass")
        t = TokenConverter(tlist)
        t.popToken()
        assert t.peekNextToken() == tlist[1]

    def testPeekNextToken_returnsZeroIfNoTokensLeft(self):
        tlist = getTokenList("class foo: pass")
        t = TokenConverter(tlist)
        for i in range(0,5):
            t.popToken()
        assert t.peekNextToken() == 0

    # a basic functional test
    def testTokenConverterWithClass(self):
        tlist = getTokenList(testdata_tokenconvertersrc)
        tc = TokenConverter(tlist)
        ast = parser.suite(testdata_tokenconvertersrc)
        astl = ast.tolist(1)
        classastl = astl[1][1][1]
        toklist = tc.getTokens(classastl)
        # toklist contains everything apart from the last ENDMARKER token
        assert(toklist == tlist[0:-1])

    # check it works with tuples too!
    def testTokenConverterWithTuple(self):
        tlist = getTokenList(testdata_tokenconvertersrc)
        tc = TokenConverter(tlist)
        ast = parser.suite(testdata_tokenconvertersrc)
        astl = ast.totuple(1)
        classastl = astl[1][1][1]
        toklist = tc.getTokens(classastl)
        # toklist contains everything apart from the last ENDMARKER token
        assert(toklist == tlist[0:-1])



class TestGetDedent(unittest.TestCase):

    class MockDedentToken:
        def __init__(self):
            self.type = tokenize.DEDENT

    class MockNewlineToken:
        def __init__(self):
            self.type = tokenize.NEWLINE

    class MockNLToken:
        def __init__(self):
            self.type = tokenize.NL
    
    def testGetDedentIfPresent_returnsDedentIfPresent(self):
        dedent = TestGetDedent.MockDedentToken()
        tc = TokenConverter([dedent])
        tokens = tc.getDedentIfPresent()
        assert(tokens == [dedent])

    def testGetDedentIfPresent_returnsEmptyListIfDedentNotPresent(self):
        newline = TestGetDedent.MockNewlineToken()
        tc = TokenConverter([newline])
        tokens = tc.getDedentIfPresent()
        assert(tokens == [])

    def testGetDedent_IncludesTrailingNewlinesAndNLs(self):
        dedent = TestGetDedent.MockDedentToken()
        newline = TestGetDedent.MockNewlineToken()
        nl = TestGetDedent.MockNLToken()
        tc = TokenConverter([dedent,newline,nl,nl])
        tokens = tc.getDedentIfPresent()
        assert(tokens == [dedent,newline,nl,nl])


    def testGetDedent_returnsDedentIfPresent(self):
        dedent = TestGetDedent.MockDedentToken()
        tc = TokenConverter([dedent])
        tokens = tc.getDedentIfPresent()
        assert(tokens == [dedent])
        
    def testGetDedent_throwsExceptionIfNoDedentPresent(self):
        newline = TestGetDedent.MockNewlineToken()
        tc = TokenConverter([newline])
        try:
            tokens = tc.getDedent()
        except:
            return
        self.fail("exception should have been raised")




class TestNodeIsToken(unittest.TestCase):

    tokenNode = [tokenize.NAME, 'class', 1]
    notTokenNode = [250, '', 1]  

    def testNodeIsToken_returnsTrueForToken(self):
        tc = TokenConverter([])
        assert(tc._nodeIsToken(TestNodeIsToken.tokenNode))

    def testNodeIsToken_returnsFalseForNonToken(self):
        tc = TokenConverter([])
        assert(tc._nodeIsToken(TestNodeIsToken.notTokenNode) == 0)


def getParseTree(src):
    return parser.suite(src).tolist(1)

class TestGetFirstTokenNode(unittest.TestCase):

    def testGetFirstTokenNode_SkipsNonTokenNodesAndReturnsTokenNode(self):
        src = testdata.pass_nocomment[1:]
        tc = getTokenConverter(src)
        parsetree = getParseTree(src)
        assert tc._getFirstTokenNode(parsetree) == [1,'pass',1]


class TestGetTokensStartBlock(unittest.TestCase):

    def testGetTokensStartBlock_IncludesDocString(self):
        src = testdata_class_with_docstring
        tlist = getTokenList(src)
        tc = TokenConverter(tlist)
        toks = tc.getTokensStartBlock()
        # ensure that it has the string and newline token
        assert(toks[-2].type == tokenize.STRING)
        assert(toks[-1].type == tokenize.NEWLINE)




testdata_tokenconvertersrc = """
class foo:
    def bah(self): # my comment
        pass
"""



testdata_class_with_docstring = """
class foo:
    \"\"\"docstring\"\"\"
    pass
"""

testdata_passWithSingleLineComment = """
pass #comments pass
#comments Next code block
pass
"""

testdata_passWithMultiLineComment = """
pass #comments pass
#comments Next code block
pass
"""


class TestGetCommentAndNewline(unittest.TestCase):

    def testGetCommentAndNewline_returnsEmptyListIfNextTokenNotCommentOrNewline(self):
        src = testdata.pass_nocomment[1:]
        tc = TokenConverter(getTokenList(src)[2:])
        assert tc.getCommentAndNewline() == []

    def testGetCommentAndNewline_returnsSingleLineCommentButNotNextCodeToken(self):
        src = testdata.pass_withcomment[1:]
        tc = TokenConverter(getTokenList(src))
        assert tc.getCommentAndNewline() == getTokenList(src)[0:1]

    def testGetCommentAndNewline_returnsInlineCommentAndNewline(self):
        src = testdata.pass_inlinecomment[1:]
        tc = TokenConverter(getTokenList(src)[1:])
        assert tc.getCommentAndNewline() == getTokenList(src)[1:3]

    def testGetCommentAndNewline_returnsMultiLineCommentButNotNextCodeToken(self):
        src = testdata.pass_withpreceedingmultilinecomment[1:]
        tc = TokenConverter(getTokenList(src))
        assert tc.getCommentAndNewline() == getTokenList(src)[0:2]
        
    def testGetCommentAndNewline_returnsMultiLineCommentButNotCommentForNextCodeToken(self):
        src = testdata.pass_withMultilineComment[1:]
        tc = TokenConverter(getTokenList(src)[1:])
        assert tc.getCommentAndNewline() == getTokenList(src)[1:5]
        

    def testGetCommentAndNewline_returnsMultiLineCommentButNotSucceedingBlankLine(self):
        src = testdata.pass_withMultilineCommentAndBlankLine[1:]
        tc = TokenConverter(getTokenList(src)[1:])
        assert tc.getCommentAndNewline() == getTokenList(src)[1:5]


    def testGetCommentAndNewline_returnsNewlineButNotCommentForNextCodeToken(self):
        src = testdata.pass_commentonnextline[1:]
        tc = TokenConverter(getTokenList(src)[1:])
        assert tc.getCommentAndNewline() == getTokenList(src)[1:2]


class TestGetTokensPreceedingNodes(unittest.TestCase):
    def testGetTokensPreceedingNodes_returnsEmptyListIfNoTokensPreceedingNode(self):
        src = testdata.pass_nocomment[1:]
        parsetree = getParseTree(src)
        tc = TokenConverter(getTokenList(src))
        assert tc.getTokensPreceedingNodes(parsetree) == [] 
        
    def testGetTokensPreceedingNodes_returnsCommentPreceedingPassToken(self):
        src = testdata.pass_withcomment[1:]
        parsetree = getParseTree(src)
        tc = TokenConverter(getTokenList(src))
        assert tc.getTokensPreceedingNodes(parsetree) == [getTokenList(src)[0]]
        
    def testGetTokensPreceedingNodes_worksWhenParsetreesAreVirtuallyIdentical(self):
        src = testdata_getAttrTwice[1:]
        parsetree = getParseTree(src)        
        tc = TokenConverter(getTokenList(src))
        # previous tokens are a.b   (parsetree[2] is a.c)
        assert tc.getTokensPreceedingNodes(parsetree[2]) == getTokenList(src)[:4]

    
    def testGetTokenNodes_returnsListOfNodesWhichCorrespondToTokens(self):
        src = "a.b"
        parsetree = getParseTree(src)        
        tc = TokenConverter(getTokenList(src))
        assert tc._getTokenNodes(parsetree) ==[[1, 'a', 1], [23, '.', 1], [1, 'b', 1], [4, '', 1], [0, '', 1]]

    def testGetOffsetOfTokensCorrespondingToParseTree_returnsCorrectTokenList(self):
        src = testdata_getAttrTwice[1:]
        parsetree = getParseTree(src)
        tc = TokenConverter(getTokenList(src))
        assert tc._getOffsetOfTokensCorrespondingToParseTree(parsetree[2]) == 4


testdata_getAttrTwice = """
a.b
a.c
"""

if __name__ == "__main__":
    unittest.main()
    #import pdb
    #pdb.run('unittest.main()')
